-------------------------------------------------------------------------------
--
-- Copyright 2022
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------
-- Author: Job van Wee
-- Purpose: resize the data into a single sosi with a width of g_out_data_w.
--
-- Description:
--  The data gets collected into a vector(c_v) and from this vector the output
--  data gets read. When the reading passes the halfway point of c_v new data
--  is requested and the whole vector shifts g_in_data_w amount of bits to fit
--  the new data at the end.
--
-- Remark:
--  Use VHDL coding template from:
--  https://support.astron.nl/confluence/display/SBe/VHDL+design+patterns+for+RTL+coding
--  The output vector must be larger than the input vector.

LIBRARY IEEE, dp_lib, tech_ddr_lib, common_lib;
USE IEEE.std_logic_1164.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE common_lib.common_pkg.ALL;

ENTITY ddrctrl_output_unpack IS
  GENERIC (
    g_tech_ddr      : t_c_tech_ddr;
    g_in_data_w     : NATURAL;
    g_out_data_w    : NATURAL;
    g_block_size    : NATURAL
  );
  PORT (
    clk             : IN  STD_LOGIC;
    rst             : IN  STD_LOGIC;
    in_sosi         : IN  t_dp_sosi   := c_dp_sosi_init;
    in_ds           : IN  NATURAL;
    in_bsn          : IN  STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0);
    out_sosi        : OUT t_dp_sosi   := c_dp_sosi_init;
    out_ready       : OUT STD_LOGIC   := '0'
  );
END ddrctrl_output_unpack;


ARCHITECTURE rtl OF ddrctrl_output_unpack IS

  -- type for statemachine
  TYPE t_state IS ( READING, FIRST_READ, SECOND_READ, OVER_HALF, RESET, IDLE, OFF);

  -- record for readability
  TYPE t_reg IS RECORD
  state           : t_state;
  a_of            : NATURAL;
  op_data_cnt     : NATURAL;
  delay_data      : STD_LOGIC_VECTOR(g_in_data_w-1 DOWNTO 0);
  dd_fresh        : STD_LOGIC;
  c_v             : STD_LOGIC_VECTOR(g_in_data_w*2-1 DOWNTO 0);
  sr_done         : STD_LOGIC;
  bsn_cnt         : NATURAL;
  out_sosi        : t_dp_sosi;
  out_ready       : STD_LOGIC;
  END RECORD;

  CONSTANT c_t_reg_init   : t_reg     := (RESET, 0, 0, (OTHERS => '0'), '0', (OTHERS => '0'), '0', 0, c_dp_sosi_init, '0');


  -- signals for readability
  SIGNAL d_reg            : t_reg     := c_t_reg_init;
  SIGNAL q_reg            : t_reg     := c_t_reg_init;

BEGIN

  q_reg <= d_reg WHEN rising_edge(clk);

  -- put the input data into c_v and fill the output vector from c_v
  p_state : PROCESS(q_reg, rst, in_sosi, in_ds, in_bsn)

    VARIABLE v            : t_reg;

  BEGIN

    v := q_reg;

    CASE q_reg.state IS
    WHEN READING =>
      -- generating output from the data already present in c_v
      v.out_ready := '0';
      v.out_sosi.data(g_out_data_w-1 DOWNTO 0) := q_reg.c_v((g_out_data_w*(q_reg.op_data_cnt+1))+q_reg.a_of-1 DOWNTO (g_out_data_w*q_reg.op_data_cnt)+q_reg.a_of);
      v.out_sosi.valid := '1';
      v.bsn_cnt := q_reg.bsn_cnt+1;
      v.op_data_cnt := q_reg.op_data_cnt+1;

      IF q_reg.bsn_cnt = g_block_size-1 THEN
        v.out_sosi.bsn(c_dp_stream_bsn_w-1 DOWNTO 0) := INCR_UVEC(q_reg.out_sosi.bsn, 1);
        v.bsn_cnt := 0;
      END IF;

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;


      IF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '1' THEN
        v.state := OVER_HALF;
      ELSIF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '0' THEN
        v.state := IDLE;
      ELSE
        v.state := READING;
      END IF;



    WHEN OVER_HALF =>
      -- generating output data from c_v but past the halfway point of c_v so there needs to be new data added
      v.out_ready := '1';
      v.out_sosi.data(g_out_data_w-1 DOWNTO 0) := q_reg.c_v((g_out_data_w*(q_reg.op_data_cnt+1))+q_reg.a_of-1 DOWNTO (g_out_data_w*q_reg.op_data_cnt)+q_reg.a_of);
      v.out_sosi.valid := '1';
      v.bsn_cnt := q_reg.bsn_cnt+1;
      v.c_v(g_in_data_w-1 DOWNTO 0) := q_reg.c_v(g_in_data_w*2-1 DOWNTO g_in_data_w);
      v.c_v(g_in_data_w*2-1 DOWNTO g_in_data_w) := q_reg.delay_data(g_in_data_w-1 DOWNTO 0);
      v.dd_fresh := '0';
      v.a_of := (g_out_data_w*(q_reg.op_data_cnt+1))+q_reg.a_of-g_in_data_w;
      v.op_data_cnt := 0;

      IF q_reg.bsn_cnt = g_block_size-1 THEN
        v.out_sosi.bsn(c_dp_stream_bsn_w-1 DOWNTO 0) := INCR_UVEC(q_reg.out_sosi.bsn, 1);
        v.bsn_cnt := 0;
      END IF;

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;


      IF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '1' THEN
        v.state := OVER_HALF;
      ELSIF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '0' THEN
        v.state := IDLE;
      ELSE
        v.state := READING;
      END IF;



    WHEN FIRST_READ =>
      -- fills the first half of c_v and generates a output from it.
      v.out_ready := '0';
      v.c_v(g_in_data_w-1 DOWNTO 0) := q_reg.delay_data(g_in_data_w-1 DOWNTO 0);
      v.dd_fresh := '0';
      v.a_of := in_ds+2+((5)*14);  -- will be fixed after in_ds is fixed in ddrctrl_controller.
      v.out_sosi.data(g_out_data_w-1 DOWNTO 0) := v.c_v(g_out_data_w+v.a_of-1 DOWNTO v.a_of);
      v.out_sosi.valid := '1';
      v.out_sosi.bsn(c_dp_stream_bsn_w-1 DOWNTO 0) := in_bsn(c_dp_stream_bsn_w-1 DOWNTO 0);

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;


      IF v.dd_fresh = '1' THEN
        v.state := SECOND_READ;
      ELSE
        v.state := IDLE;
      END IF;

    WHEN SECOND_READ =>
      -- fills the second half of c_v and generates a output from it.
      v.out_ready       := '0';
      v.c_v(g_in_data_w*2-1 DOWNTO g_in_data_w) := q_reg.delay_data(g_in_data_w-1 DOWNTO 0);
      v.dd_fresh        := '0';
      v.out_sosi.data(g_out_data_w-1 DOWNTO 0) := v.c_v(g_out_data_w*2+q_reg.a_of-1 DOWNTO g_out_data_w+q_reg.a_of);
      v.out_sosi.valid  := '1';
      v.bsn_cnt         := 1;
      v.op_data_cnt     := 2;
      v.sr_done         := '1';

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;


      IF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '1' THEN
        v.state := OVER_HALF;
      ELSIF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '0' THEN
        v.state := IDLE;
      ELSE
        v.state := READING;
      END IF;


    WHEN RESET =>
      v := c_t_reg_init;

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;
      v.state := OFF;


    WHEN IDLE =>
      -- the statemachine goes to Idle when its finished or when its waiting on other components.
      v.out_ready := '1';
      v.out_sosi.valid := '0';

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;


      IF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '1' AND q_reg.sr_done = '1' THEN
        v.state := OVER_HALF;
      ELSIF (g_out_data_w*(v.op_data_cnt+1))+q_reg.a_of >= g_in_data_w AND v.dd_fresh = '1' AND q_reg.sr_done = '0' THEN
        v.state := SECOND_READ;
      ELSE
        v.state := IDLE;
      END IF;



    WHEN OFF =>
      -- the stamachine has a state off so it knows when to go to first read, it can't go to first read from IDLE
      v.out_ready := '1';
      v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);

      IF in_sosi.valid = '1' THEN
        v.delay_data(g_in_data_w-1 DOWNTO 0) := in_sosi.data(g_in_data_w-1 DOWNTO 0);
        v.dd_fresh := '1';
      END IF;


      IF in_sosi.valid = '1' THEN
        v.state := FIRST_READ;
      ELSE
        v.state := OFF;
      END IF;



    END CASE;

    IF rst = '1' THEN
      v.state := RESET;
    END IF;
    d_reg <= v;
  END PROCESS;

  -- fill outputs
  out_sosi <= q_reg.out_sosi;
  out_ready <= q_reg.out_ready;


END rtl;
